#if 0
  Siril 
  Copyright (C) 2005 Francois Meyer (dulle @ free.fr)
  Reference site is http://dulle.free.fr/alidade/siril.php

  Siril is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#endif
#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <fitsio.h>
#include <complex.h>
#include <fftw3.h>
#include "callbacks.h"
#include "siril.h"
#include "Def_Wavelet.h"
#define DEFINC 4
#define DEFDEC 20 

char error_buffer[256];
fits dark, offset, flat ;

unsigned base, nbdata;
int debug;

unsigned char isrgb(fits *fit){
	if (fit->naxis==3){
		return 1;
		}
	else {
		return 0;
		}
	}

int readfits(char *filename, fits *fit){
	int status=0;
	long orig[3]={1L,1L,1L};
	int zero=0;
	char name[255]="/tmp/tmp_siril.fit";

	if(stat_file(filename,&(com.imagetype))){
		snprintf(error_buffer,255,"readfile: %s.[fit|fits|png|jpeg|jpg|bmp] not found\n", filename);
		siril_error(error_buffer,0);
		return 1;
		}

	strncpy(name,filename,255);

	status=0;
	if ( fits_open_file(&(fit->fptr), name, READONLY, &status) )
	report_fits_error( status , error_buffer);
	if (status){
		snprintf(error_buffer,256,"fitsio error, file %s\n",filename);
		siril_error(error_buffer,0);
		return status;
		}

	status=0;
	fits_get_img_param(fit->fptr, 2, &(fit->bitpix), &(fit->naxis), fit->naxes, &status);
	if (status){
		snprintf(error_buffer,256,"fitsio error, file %s\n",filename);
		siril_error(error_buffer,0);
		report_fits_error( status , error_buffer);
		return status;
		}

	fit->rx=fit->naxes[0];
	fit->ry=fit->naxes[1];
	nbdata=fit->rx*fit->ry;

	if(fit->naxis==3){
		fit->naxes[2]=3;
		}
	else{
		fit->naxes[2]=1;
		}
	if((fit->data=(short *)realloc(fit->data,nbdata*fit->naxes[2]*sizeof(WORD)))==NULL){
		fprintf(stderr,"readfits: error realloc %s %lu\n",filename, nbdata*fit->naxes[2]);
		return -1;
		}

	if(fit->naxis==3){
		fit->pdata[RLAYER]=fit->data;
		fit->pdata[GLAYER]=fit->data+nbdata;
		fit->pdata[BLAYER]=fit->data+nbdata*2;
		if(fit==&gfit){
			rgbprocess(TRUE);	
			}
		}
	else{
		fit->pdata[RLAYER]=fit->data;
		fit->pdata[GLAYER]=fit->data;
		fit->pdata[BLAYER]=fit->data;
		}
	
	status=0;
	fits_read_pix(fit->fptr,TSHORT,orig,fit->naxes[0]*fit->naxes[1]*fit->naxes[2], &zero,fit->data,&zero,&status);
	if (status){
		snprintf(error_buffer,256,"fitsio error, file %s\n",filename);
		siril_error(error_buffer,0);
		report_fits_error( status, error_buffer );
		}
	status=0;
	fit->bitpix=SIZESHORT;

/*	fit->hi=0;
	fit->lo=0;*/
	fits_read_key (fit->fptr,TUSHORT, "MIPS-HI", &(fit->hi), NULL, &status);
	fits_read_key (fit->fptr,TUSHORT, "MIPS-LO", &(fit->lo), NULL, &status);
	fits_close_file(fit->fptr,&status);	
	snprintf(error_buffer,256,"file %s, %ld layers\n",filename, fit->naxes[2]);
	siril_error(error_buffer,0);
	return 0;
	}

int readbmp(char *name, fits *fit){
	BYTE header[256];
	int fd, nbplane;
	int count, i , lx, ly;
	unsigned char *buf;
	
	fd=open(name,O_RDONLY);
	if (fd==-1){
		perror("Fichier rat");
		return 0;
		}

	if((count=read(fd, header,54))!=54){
		fprintf(stderr,"readbmp: %d octets lus\n",count);
		perror("readbmp: rat la lecture du header");
		return 0;
		}

	lx=(int)(256*header[19]+header[18]);
	ly=(int)(256*header[23]+header[22]);
	nbplane=header[28]/8;
	nbdata=(lx)*(ly)*nbplane;
	if (nbplane==1){
		buf=(unsigned char *)malloc((lx) * (ly)*nbplane+1024);
		if((i=read(fd,buf,1024))!=1024){
			fprintf(stderr,"readbmp: %d octets lus\n",i);
			perror("readbmp: rat la lecture de la lut");
			
			}
		}
	else{
		buf=(unsigned char *)malloc((lx) * (ly)*nbplane);
		}

	if((i=read(fd,buf,nbdata))!=nbdata){
		perror("readbmp: %d lus, %d attendus\n"); 
		};

	close(fd);

//	reallocgfit(*lx,*ly);
	if(nbplane==3){
		fprintf(stderr,"readbmp: bmp24tofits\n");
		bmp24tofits(buf, lx, ly, fit);
		}
	if (nbplane==1){
		fprintf(stderr,"readbmp: bmp8tofits\n");
		bmp8tofits(buf, lx, ly, fit);
		}
	free(buf);
	return nbplane;
	}

int savefits(char *name, fits *f){
	int status;
	long orig[3]={1L,1L,1L};
	char buffer[256];

	f->naxes[0]=f->rx;
	f->naxes[1]=f->ry;

	if(f->naxis==3){
		f->naxes[2]=3;
		}
	if(f->naxis==2){
		f->naxes[2]=1;
		}

	if(strstr(name,".fit")==NULL){
		snprintf(buffer,256,"%s.fit",name);
		}
	else{
		snprintf(buffer,255,name);	
		}

	remove(buffer);               /* Delete old file if it already exists */
	status=0;
	if (fits_create_file(&(f->fptr), buffer, &status)) /* create new FITS file */
			report_fits_error( status , error_buffer);           /* call printerror if error occurs */
	if ( fits_create_img(f->fptr,  f->bitpix, f->naxis, f->naxes, &status) )
				 report_fits_error( status , error_buffer);          

	if(f->bitpix==16){
		if ( fits_write_pix(f->fptr, TSHORT, orig, 
				f->naxes[0] * f->naxes[1] *f->naxes[2], f->data, &status) ){
			report_fits_error( status , error_buffer);          
			}
		}

	if(f->bitpix==8){
		if ( fits_write_pix(f->fptr, TBYTE, orig, 
				f->naxes[0] * f->naxes[1] *f->naxes[2], f->cdata, &status) ){
			report_fits_error( status , error_buffer);          
			}
		}

	fits_update_key (f->fptr, TUSHORT, "MIPS-HI", &(f->hi),
                "Upper visualization cutoff ", &status);
	fits_update_key (f->fptr, TUSHORT, "MIPS-LO", &(f->lo),
                "Lower visualization cutoff ", &status);
	fits_close_file(f->fptr,&status);
	return 0;
	}

int savebmp(char *name, fits *fit, int rx, int ry){
	int fd, size, i, j, seuil=0, div;
	unsigned char rgb;
	unsigned char *buf;
	short *rbuf, *gbuf, *bbuf;
//			     0                          10        14
	char headerbmp[54]={0x42,0x4d,0,0,0,0,0,0,0,0,0x36,0,0,0,0x28,0,
//				16                        28
				0, 0,0,0,0,0,0,0,0,0,1,0,0x18,0,0,0};

	rgb=isrgb(fit);
	nbdata=rx*ry*fit->naxes[3];
	headerbmp[28]=8*fit->naxes[3];	
	buf=malloc(nbdata);

	size=(54+rx*ry*3)%65536;
	headerbmp[2]=size%256;
	headerbmp[3]=size/256;
	size=(54+rx*ry*3)/65536;
	headerbmp[4]=size%256;
	headerbmp[5]=size/256;
	
	size=(rx*ry*3)%65536;
	headerbmp[34]=size%256;
	headerbmp[35]=size/256;
	size=(rx*ry*3)/65536;
	headerbmp[36]=size%256;
	headerbmp[37]=size/256;
	
	headerbmp[18]=rx%256;
	headerbmp[19]=rx/256;
	headerbmp[22]=ry%256;
	headerbmp[23]=ry/256;

	rbuf=fit->pdata[RLAYER];
	gbuf=fit->pdata[GLAYER];
	bbuf=fit->pdata[BLAYER];
	for (i=0; i<fit->rx; ++i){
		for (j=0; j<fit->ry; ++j){
			seuil=max(seuil,*rbuf++);
			if(rgb){
				seuil=max(seuil,*gbuf++);
				seuil=max(seuil,*bbuf++);
				}
			}	
		}
 	div=255./(double)seuil;
	rbuf=fit->pdata[RLAYER];
	gbuf=fit->pdata[GLAYER];
	bbuf=fit->pdata[BLAYER];
	
	for (i=0; i<fit->rx; ++i){
		for (j=0; j<fit->ry; ++j){
			*buf++=*rbuf++*div;
			if(rgb){
				*buf++=*gbuf++*div;
				*buf++=*bbuf++*div;
				}
			}
		}
	
	fd=open(name,O_RDWR|O_CREAT,0644);
	size=write(fd,headerbmp,54);
	size+=write(fd,buf,nbdata);
	fprintf(stderr, "savebmp: rx %u ry %u %d\n", rx, ry, nbdata);
	close(fd);
	free(buf);
	return 0;
	}

int copyfits(fits *from, fits *to, unsigned char init, int layer){
	int depth;
	
	nbdata=from->rx * from->ry;
	if ((init & CP_EXPAND)){
		depth=3;
		}
	else {
		depth=from->naxes[2];
		}

	if ((init & CP_ALLOC)){
		if((to->data=(short *)realloc(to->data,nbdata*depth*sizeof(WORD)))==NULL){
			fprintf(stderr,"copyfits: error realloc\n");
			return -1;
			}
		to->pdata[RLAYER]=to->data;
		if (depth==3){
				to->pdata[GLAYER]=to->data+nbdata;
				to->pdata[BLAYER]=to->data+2*nbdata;
			}
		else {
				to->pdata[GLAYER]=to->data;
				to->pdata[BLAYER]=to->data;
			}
		}
//		memcpy(to->r,from->r,from->rx * from->ry*sizeof(WORD));

	if ((init & CP_COPYA)){
		memcpy(to->data,from->data,nbdata*depth*sizeof(WORD));
		}

	if((init & CP_INIT)){
		memset(to->data,0,nbdata*depth*sizeof(WORD));
		}

	if((init & CP_FORMAT)){
		to->rx=from->rx;
		to->ry=from->ry;
		to->lo=from->lo;
		to->hi=from->hi;
		to->bitpix=from->bitpix;
		to->naxis=from->naxis;
		to->naxes[0]=from->naxes[0];
		to->naxes[1]=from->naxes[1];
		to->naxes[2]=depth;
		}

	if((init & CP_EXTRACT)){
		to->rx=from->rx;
		to->ry=from->ry;
		to->lo=from->lo;
		to->hi=from->hi;
		to->bitpix=from->bitpix;
		to->naxis=2;
		to->naxes[0]=from->naxes[0];
		to->naxes[1]=from->naxes[1];
		to->naxes[2]=1;
		memcpy(to->data,from->pdata[layer],nbdata*to->naxes[2]*sizeof(WORD));
		}
	return 0;
	}

int fits48to3fits16(fits *rgb, fits *r, fits *g, fits *b){

	return 0;
	}

int minmax(fits *fit){
	int i, j, end, layer;
	short *buf[3];
	
	if (fit->naxis==3){
		end=BLAYER;
		}
	else{
		end=RLAYER;
		}
	
	for (layer=RLAYER;layer<=end;++layer){
		buf[layer]=fit->pdata[layer];
		fit->max[layer]=0;
		fit->min[layer]=32768;
		}

	fit->maxi=0;
	fit->mini=32768;
	for (i=0;i<fit->rx;++i){
		for (j=0;j<fit->ry;++j){
			for (layer=RLAYER;layer<=end;++layer){
				fit->max[layer] = max(fit->max[layer],*buf[layer]);
				fit->min[layer] = min(fit->min[layer],*buf[layer]);
				buf[layer]++;
				}
			}
		}
	fit->maxi= max(fit->max[GLAYER], fit->max[BLAYER] );
	fit->maxi=max( fit->max[RLAYER], fit->maxi );
	fit->mini=min( fit->min[RLAYER], min(fit->min[GLAYER], fit->min[BLAYER] ));
	fprintf(stderr,"minmax %d %d %d %d %d %d %d\n",  fit->rx, fit->ry,fit->max[RLAYER], fit->max[GLAYER], fit->max[BLAYER], fit->maxi, fit->mini);
	return 0;
	}

void setlayers(void){
	if(com.rgb){
		com.beg=RLAYER;
		com.end=BLAYER;
		}
	else {
		com.beg=com.end=com.cvport;
		}
	}

int threshlo(fits *fit, int level){
	int i;
	guchar layer;
	short *buf=fit->data;

	setlayers();
	for(layer=com.beg;layer<=com.end;++layer){
		buf=fit->pdata[layer];
		for (i=0;i<fit->rx*fit->ry;++i){
			*buf= max(level,*buf);
			buf++;
			}
		}
	return 0;
	}

int threshhi(fits *fit, int level){
	int i;
       	guchar layer;
	short *buf=fit->data;

	setlayers();
	for(layer=com.beg;layer<com.end;++layer){
		buf=fit->pdata[layer];
		for (i=0;i<fit->rx*fit->ry;++i){
			*buf=min(level,*buf);
			buf++;
			}
		}
	return 0;
	}

int nozero(fits *fit, int level){
	int i;
	guchar layer;
	short *buf;
	
	setlayers();
	for(layer=com.beg;layer<=com.end;++layer){
		buf=fit->pdata[layer];
		for (i=0;i<fit->rx*fit->ry;++i){
			if(*buf==0)
				*buf=level;
			buf++;
			}
		}
	return 0;
	}

int level_adjust(fits* fit){
	if(!com.leveldrag){
		minmax(fit);
		fit->hi=fit->max[com.cvport];
		fit->lo=fit->min[com.cvport];
		}
	set_ranges();
	return 0;
	}

int soper(fits *a, float scalar, char oper){
	WORD *gbuf;
	int i, j;
	guchar layer;

	setlayers();
	for(layer=com.beg;layer<=com.end;++layer){
		gbuf=a->pdata[layer];
		switch (oper){
			case 'a':
				for (i=0;i<a->rx;++i){
					for (j=0;j<a->ry;++j){
						*(gbuf++) += scalar;
						}
					}
				break;
			case 's':
				for (i=0;i<a->rx;++i){
					for (j=0;j<a->ry;++j){
						*(gbuf++) -= scalar;
						}
					}
				break;
			case 'm':
				for (i=0;i<a->rx;++i){
					for (j=0;j<a->ry;++j){
						*(gbuf++) *= scalar;
						}
					}
				break;
			case 'd':
				for (i=0;i<a->rx;++i){
					for (j=0;j<a->ry;++j){
						*(gbuf++) /= scalar;
						}
					}
				break;
			}
		}
	return 0;
	}

int imoper(fits *a, fits *b, char oper){
	WORD *gbuf, *buf;
	int i, j;
	guchar layer;
	
	if (a->rx!=b->rx || a->ry!=b->ry){
		fprintf(stderr, "Wrong size: %d=%d? / %d=%d?\n",a->rx,b->rx,a->ry,b->ry);
		}
	else {
		setlayers();
		for(layer=com.beg;layer<=com.end;++layer){
			buf=b->pdata[layer];
			gbuf=a->pdata[layer];
			switch (oper){
				case 'a':
					for (i=0;i<b->rx;++i){
						for (j=0;j<b->ry;++j){
							*(gbuf++) += *(buf++);
							}
						}
					break;
				case 's':
					for (i=0;i<b->rx;++i){
						for (j=0;j<b->ry;++j){
							if (*buf<=*gbuf){
								*(gbuf++) -= *(buf++);
								}
							else {
								*(gbuf++) = 0;buf++;
								}
							}
						}
					break;
				case 'm':
					for (i=0;i<b->rx;++i){
						for (j=0;j<b->ry;++j){
							*(gbuf++) *= *(buf++);
							}
						}
					break;
				case 'd':
					for (i=0;i<b->rx;++i){
						for (j=0;j<b->ry;++j){
							*(gbuf++) /= *(buf++);
							}
						}
					break;
				}
			}
		}
	return 0;
	}

int fdiv(fits *a, fits *b, float coef){
	WORD *gbuf, *buf;
	int i, j;
	guchar layer; 
	unsigned temp;

	if (a->rx!=b->rx || a->ry!=b->ry){
		fprintf(stderr, "Wrong size: %d=%d? / %d=%d?\n",a->rx,b->rx,a->ry,b->ry);
		}
	else {
		setlayers();
		for(layer=com.beg;layer<=com.end;++layer){
			buf=b->pdata[layer];
			gbuf=a->pdata[layer];
				for (i=0;i<b->rx;++i){
					for (j=0;j<b->ry;++j){
						temp=(unsigned)((double)coef*((double)*gbuf/(double)*buf++));
						*(gbuf++)=temp;
						}
					}
			}
		}
	return 0;
	}

double gaussienne(double sigma, int size, double *gauss){
	double s2, s, *igauss;
	int i,j,np1,n;
	
	n=size/2;
	np1=n+1;
	s2=sigma*sigma;
	s=(double)0;
	igauss=gauss;
	//fprintf(stderr,"size: %d sigma:%f\n", size, sigma);
	for (i=0;i<size;++i){
		for (j=0;j<size;++j){
			s+=*gauss++=exp(-((double)(i-n)*(i-n) + (double)(j-n)*(j-n))
			      /2/s2);
			//fprintf(stderr,"%d:%d %f \n", i, j, igauss[i*size+j]);
			}
		}	
	//fprintf(stderr,"%f \n", s);
	return s;
	}

int unsharp(fits *fit,double sigma, double mult){
// if fabs(mult) > 0.01, unsharp computes the unsharp mask 
// else unsharp computes a gaussian filter

	double normalize,g, *gauss, *gaussbuf;
	WORD *buf, *gbuf;
	int size, ss2, stride, i, j, k, l;
	guchar layer;
	
//	fprintf(stderr,"gfitrx: %d mult :%f\n", gfit.rx, mult);
	size=(int)(4*sigma);
	if (!(size%2))
		++size;

	ss2=size/2;
	gauss=malloc(size*size*sizeof(double));
	if (!gauss){
		perror("unsharp.c: Alloc error for gauss");
	        return -1;
		}	
//	fprintf(stderr,"size: %d sigma:%f\n", size, sigma);
	
	normalize=gaussienne(sigma, size, gauss);
//	fprintf(stderr,"gfitrx: %d ss2:%d\n", gfit.rx, ss2);
	wfit[4].data=(WORD *)malloc(fit->rx*fit->ry*sizeof(WORD));

	wfit[4].rx=fit->rx;
	wfit[4].ry=fit->ry;
	wfit[4].lo=fit->lo;
	wfit[4].hi=fit->hi;
	stride=wfit[4].rx-size;
	setlayers();
	for(layer=com.beg;layer<=com.end;++layer){
		memcpy(wfit[4].data, fit->pdata[layer], fit->rx*fit->ry*sizeof(WORD));
		
		buf=fit->pdata[layer]+ss2+ss2*fit->rx;
	//	fprintf(stderr,"gfitrx: %d ss2:%d\n", gfit.rx, ss2);
		for (i=ss2;i<fit->ry-ss2;++i){
			for (j=ss2;j<fit->rx-ss2;++j){
				g=(double)0;
				gbuf=wfit[4].data+(i-ss2)*fit->rx+j-ss2;
				gaussbuf=gauss;
				for (k=0;k<size;++k){
					for(l=0;l<size;++l){	
						g+=(*gbuf++)*(*gaussbuf++);	
						}
					gbuf+=stride;
					}
				*(buf++)=g/normalize;
				}
			buf+=ss2+ss2;
			}
		
		buf=fit->pdata[layer];
		gbuf=wfit[4].data;
		if (fabs(mult)>0.01){
			for (i=0;i<fit->rx;++i){
				for(j=0;j<fit->ry;++j){
					*buf = max(0,(short)(*gbuf + mult * (*gbuf - *buf)));
					buf++;
					gbuf++;
					}
				}
			}
		else {
			for (i=0;i<fit->rx;++i){
				for(j=0;j<fit->ry;++j){
					*buf = max(0,*buf);
					buf++;
					}
				}
			}
		}	
	free(gauss);
	free(wfit[4].data);
	return 0;
	}

// inplace cropping of the image in fit 
int crop(fits *fit){
	WORD *from, *to;
	int stridefrom, i, j, layer;
	int newnbdata;
	
	newnbdata=com.llx*com.lly;
	setlayers();
	for(layer=com.beg;layer<=com.end;++layer){
		from = fit->pdata[layer] + (fit->ry -com.ly - com.lly) * fit->rx + com.lx;
		fit->pdata[layer] = fit->data+layer*newnbdata;
		to = fit->pdata[layer];
		stridefrom = fit->rx - com.llx ;
		
		for (i=0; i<com.lly; ++i){
			for (j=0; j<com.llx; ++j){
				*to++ = *from++;
				}
			from+= stridefrom;
			}
		}
	fit->rx = fit->naxes[0]=com.llx ;
	fit->ry = fit->naxes[1]=com.lly ;
	return 0;
	}

// register number images
int _register(char *genname, int number, int cx, int cy, int size, int layer){
	int status;
	int i, x, y;
	int debx, deby;
	fftw_complex *ref,*img,*in,*out,*convol;
	fftw_plan p,q;
	double qual;
	int shiftx, shifty, shift;
	double norm;
	int ret;
	int plan;
	fftw_complex *dbuf,*cbuf,*ibuf,*obuf;
	short *ubuf;

	// process_register already made sure the parameters are ok 
	// Should be better to check it here, in case... well in any case in fact.
	debx=max(0,cx-size/2);
	deby=max(0,cy-size/2);

	printf("centre : %d %d, taille : %d (%d:%d) %d\n", cx, cy, size, debx, deby, number);
	
	status=0;	

//	snprintf(buffer,255,"%s1", genname);
//	ret=readfits(buffer,&(com.r));
	fprintf(stderr,"com.seq.imgparam[0].filenum %d\n",com.seq.imgparam[0].filenum);
	ret=readfits(buildfilename(genname,com.seq.imgparam[0].filenum),&(wfit[0]));
	
	if (ret){
		return ret;
		}
	
        ref = fftw_malloc(sizeof(fftw_complex) * size * size);
        img = fftw_malloc(sizeof(fftw_complex) * size * size);
        in = fftw_malloc(sizeof(fftw_complex) * size * size);
        out = fftw_malloc(sizeof(fftw_complex) * size * size);
        convol = fftw_malloc(sizeof(fftw_complex) * size * size);
        
	if (number>200){
		plan=FFTW_MEASURE;
		}
	else {
		plan=FFTW_ESTIMATE;
		}
	p = fftw_plan_dft_2d(size, size, ref, out, FFTW_FORWARD, plan);
        q = fftw_plan_dft_2d(size, size, convol, out, FFTW_BACKWARD, plan);
        
	dbuf=ref;
	ubuf=wfit[0].pdata[layer]+debx+deby*wfit[0].rx;	// register on green layer
	for(y=0;y<size;++y){
		for(x=0;x<size;x++){
			*dbuf++=(double)*ubuf++;
			}	
		ubuf+=wfit[0].rx-size;
		}
	
	fftw_execute_dft(p,ref,in); /* repeat as needed */
	qual=quality(in,size,size,0,(unsigned short)size);
	i=1;
	shiftx=shifty=0;
	com.seq.imgparam[0].qual_rank=1;
	com.seq.imgparam[0].shiftx=0;
	com.seq.imgparam[0].shifty=0;
	com.seq.imgparam[0].quality=(int)(qual *100);
	norm=size*size*size*size;
	for (i=1;i<number;++i){
		snprintf(com.status,255,"Processing image %s%d.fit", genname,com.seq.imgparam[i].filenum);
		progress(0);
		buildfilename(genname,com.seq.imgparam[i].filenum);
		if ((ret=readfits(com.formname,&(wfit[4])))==0){
			dbuf=img;
			ubuf=wfit[4].data+debx+deby*wfit[4].rx;

			status=0;
			report_fits_error(status, error_buffer);

			for(y=0;y<size;++y){
				for(x=0;x<size;x++){
					*dbuf++=(double)*ubuf++;
					}	
				ubuf+=wfit[4].rx-size;
				}
			fftw_execute_dft(p,img,out); /* repeat as needed */
			qual=quality(out,size,size,0,(unsigned short)size);

			cbuf=convol;
			ibuf=in;
			obuf=out;
			for (x=0;x<size*size;x++){
					*cbuf++=*ibuf++*conj(*obuf++);
					}

			fftw_execute_dft(q,convol,ref); /* repeat as needed */
			shift=0;
			for (x=1;x<size*size;++x){
				if(creal(ref[x])>creal(ref[shift])){
					shift=x;
					}
				}
			shifty=shift/size;
			shiftx=shift%size;
			if (shifty>size/2){
				shifty-=size;
				}
			if (shiftx>size/2){
				shiftx-=size;
				}
			com.seq.imgparam[i].qual_rank=i+1;
			com.seq.imgparam[i].shiftx=shiftx;
			com.seq.imgparam[i].shifty=shifty;
			com.seq.imgparam[i].quality=(int)(qual*100);
			}
		else {
			return ret;
			}
		}

	status=0;
	
	fftw_destroy_plan(p);
	fftw_destroy_plan(q);
        fftw_free(in); 
	fftw_free(out);
        fftw_free(ref);
       	fftw_free(img);
       	fftw_free(convol);
		fprintf(stderr,"gr %d %d %d %d %d %d\n",  com.seq.imgparam[i].filenum, 
									com.seq.imgparam[i].qual_rank, 
									com.seq.imgparam[i].shiftx, 
									com.seq.imgparam[i].shifty, 
									com.seq.imgparam[i].quality, 
									com.seq.imgparam[i].incl); //default is to include all images, hence the zero

	return(0);
	}

int composit(char *genname, int number, char shifted){
	int x,y,nx,ny,i,ii,j,shiftx,shifty, layer, index, imnum;
	long  *somme[3], *from, maxim;
	short *to;
	FILE *sf=NULL;
	double ratio;
	fits *fit=wfit;

	index=com.seq.imgparam[0].filenum;
//	fprintf(stderr,"Processing image %s%d.fit seq %s\n", genname,1, com.seq.name);
	snprintf(com.status,255,"Processing image %s%d.fit", genname,com.seq.imgparam[0].filenum);
	progress(0);

	buildfilename(genname,index);
	readfits(com.formname,fit);
	nbdata=fit->ry*fit->rx;

	somme[0]=(long *)malloc(nbdata*sizeof(long)*fit->naxes[2]);
	if(fit->naxis==3 && com.rgb){
		somme[1]=somme[0]+nbdata;
		somme[2]=somme[0]+nbdata*2;
		}	

	if (somme==NULL){
		siril_error("composit: memory allocation failure\n",0);
		return 0;
		}
	
	memset(somme[0], nbdata*fit->naxes[2]*sizeof(long), 0);

	shiftx=shifty=0;
	maxim=0;

	for (j=0;j<number;++j){
		fprintf(stderr,"composit %d %d shifted %d\n", shiftx,shifty,shifted);
		imnum=com.seq.imgparam[j].filenum;
		if(shifted){
			shiftx=com.seq.imgparam[j].shiftx;
			shifty=com.seq.imgparam[j].shifty;
			}
		i=0;
		if(com.seq.active && (com.seq.imgparam[j].incl!=0 || imnum>com.seq.end)){
			snprintf(com.status,255,"Image %s%d.fit is excluded", genname,imnum);
			fprintf(stderr,"Image %s%d.fit is excluded\n", genname,imnum);
			progress(0);
			}
		else {
			snprintf(com.status,255,"Processing image %s%d.fit", genname,imnum);
			progress(0);

			buildfilename(genname,imnum);
			if(readfits(com.formname,fit)>0){
				return 1;
				}

			for (y=0;y<fit->ry;++y){
			  for (x=0;x<fit->rx;++x){
				nx=(x-shiftx);
				ny=(y-shifty);
								//printf("x %d y %d sx %d sy %d 	i %d ii %d\n",x,y,shiftx,shifty,i,ii);
				if (nx>=0 && nx<fit->rx && ny>=0 && ny<fit->ry){
					ii=ny*fit->rx+nx;
					//printf("%d %d 	%d %d\n",shiftx,shifty,i,ii);
					if (ii>0 && ii<fit->rx*fit->ry){
						for(layer=com.beg;layer<=com.end;++layer){
							somme[layer][i]+=fit->pdata[layer][ii];
							if (somme[layer][i]>maxim){
								maxim=somme[layer][i];
								}
							}
						}
					}
				++i;
				}
			}
		}
	}
	
	copyfits(fit, &gfit, CP_ALLOC|CP_FORMAT,0);
	for(layer=com.beg;layer<=com.end;++layer){
		if (maxim>32767){
			ratio=(double)32767/(double)maxim;
			}
		else {
			ratio=1;
			}

		from=somme[layer];
		to=gfit.pdata[layer];
		for (y=0;y<fit->ry*fit->rx;++y){
			*to++=(short)((*from++)*ratio);
			}	
		}
		
	if(sf!=NULL){
		fclose(sf);
		}
	free(somme[0]);
	return 0;
	}

int shift(int sx, int sy){
	int x, y, nx, ny, i, ii, layer;

	copyfits(&(gfit),&(wfit[5]),CP_ALLOC|CP_FORMAT,0);
	setlayers();
	i=0;
	for (y=0;y<gfit.ry;++y){
		for (x=0;x<gfit.rx;++x){
			nx=(x-sx);
			ny=(y-sy);
			//printf("x %d y %d sx %d sy %d 	i %d ii %d\n",x,y,shiftx,shifty,i,ii);
			if (nx >= 0 && nx < gfit.rx && ny >= 0 && ny < gfit.ry){
				ii=ny*gfit.rx+nx;
				//printf("%d %d 	%d %d\n",shiftx,shifty,i,ii);
				if (ii > 0 && ii < gfit.rx * gfit.ry){
					for (layer=com.beg;layer<=com.end;++layer){
						wfit[5].pdata[layer][i] = gfit.pdata[layer][ii];
						}
					}
				}
			++i;
			}
		}

	for (layer=com.beg;layer<=com.end;++layer){
		memcpy(gfit.pdata[layer], wfit[5].pdata[layer], gfit.rx * gfit.ry * sizeof(WORD));
		}
	
	return 0;
	}
	
int rshift2(char *genname, char *outname, int number, char *shiftfile){
	int i,j,shiftx,shifty,count,n;
	char line[256];
	FILE *sf=NULL;

	if (shiftfile!=NULL){
		sf=fopen(shiftfile,"r");
		if (sf==NULL){
			snprintf(error_buffer,255,"rshift2 : could not open shift file %s\n", shiftfile);
			siril_error(error_buffer,0);
			return 0;
			}
		fgets(line,255,sf);
		while(line[0]=='#'){
			fgets(line,255,sf);
			}
		}
	shiftx=shifty=0;

	for (j=1;j<=number;++j){
		if(sf!=NULL){
			count=sscanf(line,"%d %d %d",&n,&shiftx,&shifty);
			if(count!=3){
				snprintf(error_buffer,255,"rshift2 : format error in shift file %s\n", shiftfile);
				siril_error(error_buffer,0);
				return 0;
				}
			fgets(line,255,sf);
			}
		else {
			shiftx=0;
			shifty=0;
			n=j;
			}
		fprintf(stderr,"rshift2 %d %d %d\n", n, shiftx,shifty);
		i=0;
		snprintf(com.status,255,"Processing image %s%d.fit", genname,n);
		progress(0);

		buildfilename(genname,n);
		if(readfits(com.formname,&(gfit))>0){
			fprintf(stderr,"missed %s com.formanme",com.formname);
			return 1;
			}
		shift(shiftx,shifty);
		buildfilename(outname,n);
		savefits(com.formname,&gfit);
		}
	if(sf!=NULL){
		fclose(sf);
		}
	return 0;
	}
	
/* code borrowed from Lynkeos */
double quality(   fftw_complex *spectrum, unsigned int  rx, unsigned int ry,
                       unsigned short down, unsigned short up )
{
    unsigned short x, y;
    double q = 0.0;
    unsigned long d2 = down*down, u2 = up*up;
    unsigned long n = 0;
    double lum = (creal(spectrum[0]))/(double)rx/(double)ry;
    short dx , dy ;
    long f2;
    fftw_complex s;

    for( y = 0; y < ry; y++ )
    {
        for ( x = 0; x < rx/2 +1; x++ )
        {
		dx=x;
		dy=y;
            if ( dy >= rx/2 )
                dy -= rx;
            f2 = dx*dx + dy*dy;
            if ( f2 > d2 && f2 < u2 )
            {
                s = spectrum[y*(rx/2+1)+x];
                q += sqrt( creal(s) * creal(s) + cimag (s) * cimag (s) );
                n++;
//	fprintf(stderr,"5x %d y %d q %e lum %e f2 %ld d2 %lu u2 %lu\n",x,y,q, lum,f2,d2,u2);
            }
        }
    }

    return( q/lum/(double)n );
}

/* code borrowed from Lynkeos */
double entropy( short *image, unsigned int rx, unsigned int ry ) {
   // Padded width for in place transform
   double e = 0, bmax = 0;
   unsigned long x, y;

   // Compute the quadratic pixel sum
   for( y = 0; y < ry; y++ )
      for( x = 0; x < rx; x++ )
         bmax += image[y*rx+x] * image[y*rx+x];

   bmax = sqrt(bmax);

   // Compute the entropy
   for( y = 0; y < ry; y++ )
   {
      for( x = 0; x < rx; x++ )
      {
         double b = image[y*rx+x]/bmax;
         if ( b > 0.0 )
            e -= b * log(b);
      }
   }
   return( e );
}

// This is a bit special as median stacking requires
// all images to be in memory
// So we dont use the generic readfits but directly
// the cfitsio routines, and allocates as many
// pix tables as needed
//
int medstack(char * genname, int number, char *outfile) {
    fitsfile *outfptr;  /* FITS file pointers */
    fitsfile *fptr[50];  /* FITS file pointers */
    int status = 0;  /* CFITSIO status value MUST be initialized to zero! */
    int bitpix;
    int naxis[50], ii;
    long npixels = 1, firstpix[3] = {1,1,1} ;
    long naxes[50][3];    
    short *pix[50];
    int i, j, jj, type, index, beg, firstpass=1;

    if (0) { 
      printf("Usage: medstack proto number outimage \n");
      printf("\n");
      printf("Performs median stack of up to 50 images\n");
      printf("creating a new output image.  \n");
      printf("\n");
      printf("Example: \n");
      printf("  medstack in 12 out - meds in1.fit, in2.fit... in12.fit outputs to out.fit\n");
      return(0);
    }

    if (number > 50){
		fprintf(stderr,"Number must be lower than 50\n");
		return(0);
		}

	if(com.seq.active){	
		beg=com.seq.beg;
		}
	else{
		beg=0;
		}

	index=beg;
    for (i=beg; i<=number; ++i){
		snprintf(com.status,255,"Processing image %s%d.fit", genname,index);
		progress(0);
    	if(com.seq.active){
			while (com.seq.imgparam[index].incl!=0){
				++index;
				}
			}
		else{
			index=i;
			}

		if (com.seq.active && index>com.seq.end){
			number=i-1;
			fprintf(stderr,"medstack: Only %d images selected for medstack\n", number);
			break;
			}
		else{
			naxes[i][0]=naxes[i][1]=naxes[i][2]=1;
			buildfilename(genname, index+1);
			stat_file(com.formname, &type);
			fits_open_file(&fptr[i], com.formname, READONLY, &status); /* open input images */
			fits_get_img_dim(fptr[i], naxis+i, &status);  /* read dimensions */
			fits_get_img_size(fptr[i], 3, naxes[i], &status);
			fits_get_img_type(fptr[i], &bitpix, &status);
			if (status) {
				   fits_report_error(stderr, status); /* print error message */
				   return(status);
				}
			if (naxis[i] > 3) {
				   printf("Error: images with > 3 dimensions are not supported\n");
				   return (0);
				}

			if(firstpass){
				firstpass=0;
				}
			else{
				if( naxes[i-1][0]!=naxes[i][0] ||
					naxes[i-1][1]!=naxes[i][1] ||
					naxes[i-1][2]!=naxes[i][2]) {
					printf("Error: input images have different sizes at index i=%d index=%d\n", i, index);
					return (0);
					}
				}
			}
		++index;
		}

    /* create the new empty output file if the above checks are OK */
	unlink(outfile);
	if (!fits_create_file(&outfptr, outfile, &status) ) {
	      /* copy all the header keywords from first image to new output file */
	      fits_copy_header(fptr[beg], outfptr, &status);

	      npixels = naxes[beg][0];  /* no. of pixels to read in each row */

	      for (i=beg;i<=number;++i){
		pix[i] = (short *) malloc(npixels * sizeof(int)); /* mem for 1 row */
		if (pix[i] == NULL) {
			printf("Memory allocation error at pix index %d\n",i);
			return(1);
			}
		}

	      /* loop over all planes of the cube (2D images have 1 plane) */
	      for (firstpix[2] = 1; firstpix[2] <= naxes[beg][2]; firstpix[2]++)
	      {
		/* loop over all rows of the plane */
		for (firstpix[1] = 1; firstpix[1] <= naxes[beg][1]; firstpix[1]++)
		{
		  /* Read both images as doubles, regardless of actual datatype.  */
		  /* Give starting pixel coordinate and no. of pixels to read.    */
		  /* This version does not support undefined pixels in the image. */
		for (i=beg;i<=number;++i){
		status=0;
		  if (fits_read_pix(fptr[i], TSHORT, firstpix, npixels, NULL, pix[i], NULL, &status))        
		      
		      break;   /* jump out of loop on error */
		      }
		
		for (i=0;i<npixels;++i){
			for (ii=beg;ii<number;++ii){
				for (jj=ii+1;jj<=number;++jj){
					if (pix[ii][i]<pix[jj][i]){
						j=pix[ii][i];
						pix[ii][i]=pix[jj][i];
						pix[jj][i]=j;
						}
					}
				}
			}
		  fits_write_pix(outfptr, TSHORT, firstpix, npixels,
			       pix[number/2], &status); /* write new values to output image */
		}
	      }    /* end of loop over planes */

	      fits_close_file(outfptr, &status);
		for (i=beg;i<number;++i){
			free(pix[i]);
			fits_close_file(fptr[i], &status);
			}
	    }

	    if (status) {
			fits_report_error(stderr, status); /* print any error message */
			}
		return(status);
	}

int loglut(fits *fit){
	// This function maps fit with a log LUT
	guint x, y;
	guchar layer;
	gdouble normalization, temp;
	WORD *buf[3]={fit->pdata[RLAYER],fit->pdata[GLAYER],fit->pdata[BLAYER]};

	normalization=(double)32768/log(32768);

	setlayers();
	for (y = 0; y < fit->ry; y++) {
		for (x = 0; x < fit->rx; x++) {
			for(layer=com.beg;layer<=com.end;++layer){
				temp=*buf[layer]+1;
				*buf[layer]++=normalization*log(temp);
				}
			}
		}
	fit->lo=16384;
	fit->hi=32768;
	set_ranges();
	return 0;
	}
/*
 * A Fast HSL-to-RGB Transform
 * by Ken Fishkin
 * from "Graphics Gems", Academic Press, 1990
 * */
/*
 *  * given h,s,l on [0..1],
 *   * return r,g,b on [0..1]
 *    */
void hsl_to_rgb(double h,double sl,double l,double * r,double * g,double * b) {
double v;

v = (l <= 0.5) ? (l * (1.0 + sl)) : (l + sl - l * sl);
if (v <= 0)
	{
	*r = *g = *b = 0.0;
	}
else
	{
	double m;
	double sv;
	int sextant;
	double fract, vsf, mid1, mid2;

	m = l + l - v;
	sv = (v - m ) / v;
	h *= 6.0;
	sextant = h;    
	fract = h - sextant;
	vsf = v * sv * fract;
	mid1 = m + vsf;
	mid2 = v - vsf;
	switch (sextant)
		{
		case 0: *r = v; *g = mid1; *b = m; break;
		case 1: *r = mid2; *g = v; *b = m; break;
		case 2: *r = m; *g = v; *b = mid1; break;
		case 3: *r = m; *g = mid2; *b = v; break;
		case 4: *r = mid1; *g = m; *b = v; break;
		case 5: *r = v; *g = m; *b = mid2; break;
		}
	}
}
/*
 *  * RGB-HSL transforms.
 *   * Ken Fishkin, Pixar Inc., January 1989.
 *    */

/*
 *  * given r,g,b on [0 ... 1],
 *   * return (h,s,l) on [0 ... 1]
 *    */
void rgb_to_hsl (double r,double g,double b,double *h,double *s,double *l)
{
double v;
double m;
double vm;
double r2, g2, b2;

v = MAX(r,g);
v = MAX(v,b);
m = MIN(r,g);
m = MIN(m,b);

if ((*l = (m + v) / 2.0) <= 0.0) return;
if ((*s = vm = v - m) > 0.0)
	{
	*s /= (*l <= 0.5) ? (v + m ) : (2.0 - v - m) ;
	} 
else
	return;

r2 = (v - r) / vm;
g2 = (v - g) / vm;
b2 = (v - b) / vm;

if (r == v)
	*h = (g == m ? 5.0 + b2 : 1.0 - g2);
else
	if (g == v)
		*h = (b == m ? 1.0 + r2 : 3.0 - b2);
	else
		*h = (r == m ? 3.0 + g2 : 5.0 - r2);

*h /= 6;
}

int ddp(fits *a,int level, float coeff, float sigma){
	copyfits(a,&wfit[0],CP_ALLOC|CP_COPYA|CP_FORMAT,0);
	unsharp (&wfit[0],sigma,0);
	soper(&wfit[0],level,'a');
	nozero(&wfit[0],1);
	fdiv(a,&wfit[0],level);
	soper(a,coeff,'m');
	return 0;
	}

int lrgb(fits *l, fits *r, fits *g, fits *b, fits *lrgb){
//
// Combines l r g and b components into resulting lrgb
// We transform each pixel from RGB to HSI,
// then take I from the luminance l fits and
// immediately step back to RGB to the working copy
//
	guint x, y;
	gdouble rr, gg, bb, h, s, i, ps3, dps3, qps3, dpi;
	gint maxi;
	WORD *pr, *pg, *pb, *dr, *dg, *db, *pl;

//
//  some stats used to normalize
//  	
	minmax(r);
	minmax(g);
	minmax(b);
	maxi=max(r->maxi,max(g->maxi,b->maxi));
	minmax(l);
//
// initialize pointers
// 
	pr=r->data;
	pg=g->data;
	pb=b->data;
	pl=l->data;
	dr=lrgb->pdata[RLAYER];
	dg=lrgb->pdata[GLAYER];
	db=lrgb->pdata[BLAYER];
//
// some trigo constants
// we stick to h in radians, not in degrees
//
	dpi=2*M_PI;
	ps3=M_PI/3;
	dps3=2*M_PI;
	dps3=2*M_PI/3;
	qps3=4*M_PI/3;
//
// Main loop
//
	fprintf(stderr,"HSI->RGB %d %d\n", r->ry, r->rx);
	for (y = 0; y < r->ry; y++) {
		for (x = 0; x < r->rx; x++) {
//
// First normalize rgb to [0 1] 
//
			rr=(double)(*pr++)/maxi;
			gg=(double)(*pg++)/maxi;
			bb=(double)(*pb++)/maxi;

			rgb_to_hsl(rr,gg,bb,&h,&s,&i);
//
// replace luminance
//
			i=*pl++/(double)l->maxi;
//
// and back to RGB
			hsl_to_rgb(h,s,i,&rr,&gg,&bb);
//
// now denormalize and store
//
			*dr++=(short)(rr*maxi);
			*dg++=(short)(gg*maxi);
			*db++=(short)(bb*maxi);
			}
		}
	return 0;
	}

int fwhm(fits *fit){
	WORD *from, *initfrom, *maxfrom;
	int stridefrom, i, j, layer, height, relativemax,maxi;
	double l1, l2;
	int beg, end, range;
	int sum, maxsum;
	int off=0;

	setlayers();
	
	off=snprintf(error_buffer,255,"FWHM : ");

	for(layer=com.beg;layer<=com.end;++layer){

		// 
		// First search the line of maximum intensity
		//
		from = fit->pdata[layer] + (fit->ry -com.ly - com.lly) * fit->rx + com.lx;
		stridefrom = fit->rx - com.llx ;

		maxsum=-1;
		maxi=-1;
		maxfrom=from;
		height=-1;
		for (i=0; i<com.lly; ++i){
			initfrom=from;
			sum=0;
			relativemax=-1;
			for (j=0; j<com.llx; ++j){
				relativemax=max(relativemax,*from);
				sum+=*from++;
				}

			if (sum>maxsum){
				maxsum=sum;
				maxi=i;
				maxfrom=initfrom;
				height=relativemax;
				}
			from+= stridefrom;
			}

		height/=2;
		// maxi is the line of maximum intensity
		// maxfrom is the adresse of the line maxi
		// height is the half maximum intensity on maxi line
		// then search intensity peak on the given line
		//
		// now search the segment at height/2
		// let 'from' points to the beginning of the maxi line
		from=maxfrom;
		l1=(double)-1;
		l2=(double)-1;
			for (j=0; j<com.llx && *from++<height; j++);
			beg=*(maxfrom+j-1);
			end=*(maxfrom+j);
			range=end-beg;
			l1=(double)j+(double)(height-beg)/(double)range;

			for (j=l1+1; j<com.llx && *from++>height; j++);
			beg=*(maxfrom+j-1);
			end=*(maxfrom+j);
			range=end-beg;
			l2=(double)j+(double)(height-beg)/(double)range;

//		fprintf(stderr,"%d %.2lf %.2lf\n",height, l2,l1);
//		fprintf(stderr,"FWMH : height %d layer %d = %.2lf\n", height, layer, l2 -l1);
		off+=snprintf(error_buffer+off,255,"%4.2lf ", l2-l1);
		}
	snprintf(error_buffer+off,255,"\n");
	siril_error(error_buffer,0);
	return 0;
	}

int preprocess(fits *brut){
	float level;

	//FIXME What is the estimator for level ?
	level=(float)8000;

	if(com.preprostatus & (USE_OFFSET)){
		imoper(brut,&offset,'s');
		}
	
	if(com.preprostatus & (USE_DARK)){
		imoper(brut,&dark,'s');
		}

	if(com.preprostatus & (USE_FLAT)){
		fdiv(brut,&flat,level);
		}
	
	return 0;
	}

int seqpreprocess(void){
	int i;
	char filename[256];

	readfits(com.seq.darkname,&dark);
	readfits(com.seq.flatname,&flat);
	readfits(com.seq.offsetname,&offset);

	if(com.preprostatus & (USE_DARK|USE_OFFSET)){
		imoper(&dark, &offset,'s');
		}
		
	if(com.preprostatus & (USE_DARK)){
		if(com.darkopt){
			soper(&dark, com.darkopt, 'm');
			}
		}

	if(com.preprostatus & (USE_OFFSET|USE_FLAT)){
		imoper(&flat, &offset,'s');	
		}

	for(i=com.seq.beg;i<com.seq.end;++i){
		buildseqname(com.seq.selprefix,filename,i);
		readfits(filename,&(wfit[4]));
		preprocess(&(wfit[4]));
		buildseqname(com.seq.ppprefix,filename,i);
		savefits(filename,&(wfit[4]));
		}

	return 0;
	}

